@@ -21,7 +21,7 @@ from collections import abc
 
 import requests
 
-__version__ = "3.22.5"
+__version__ = "4.0.0"
 
 DEFAULT_API_URL = "http://127.0.0.1:45869/"
 HYDRUS_METADATA_ENCODING = "utf-8"
@@ -74,6 +74,7 @@ class Permission(StringableIntEnum):
     MANAGE_PAGES = 4
     MANAGE_COOKIES = 5
     MANAGE_DATABASE = 6
+    ADD_NOTES = 7
 
 
 @enum.unique
@@ -164,7 +165,7 @@ class JSONEncoder(json.JSONEncoder):
 
 
 class Client:
-    VERSION = 23
+    VERSION = 31
 
     # Access Management
     _GET_API_VERSION_PATH = "/api_version"
@@ -182,6 +183,7 @@ class Client:
 
     # Adding Tags
     _CLEAN_TAGS_PATH = "/add_tags/clean_tags"
+    _SEARCH_TAGS_PATH = "/add_tags/search_tags"
     _ADD_TAGS_PATH = "/add_tags/add_tags"
 
     # Adding URLs
@@ -190,6 +192,10 @@ class Client:
     _ADD_URL_PATH = "/add_urls/add_url"
     _ASSOCIATE_URL_PATH = "/add_urls/associate_url"
 
+    # Adding Notes
+    _SET_NOTES_PATH = "/add_notes/set_notes"
+    _DELETE_NOTES_PATH = "/add_notes/delete_notes"
+
     # Managing Cookies and HTTP Headers
     _GET_COOKIES_PATH = "/manage_cookies/get_cookies"
     _SET_COOKIES_PATH = "/manage_cookies/set_cookies"
@@ -301,25 +307,115 @@ class Client:
 
         return response.json()
 
-    def delete_files(self, hashes: abc.Iterable[str]) -> None:
-        self._api_request("POST", self._DELETE_FILES_PATH, json={"hashes": hashes})
+    def delete_files(
+        self,
+        hashes: T.Optional[abc.Iterable[str]] = None,
+        file_ids: T.Optional[abc.Iterable[int]] = None,
+        file_service_name: T.Optional[str] = None,
+        file_service_key: T.Optional[str] = None,
+        reason: T.Optional[str] = None
+    ) -> None:
+        if hashes is None and file_ids is None:
+            raise ValueError("At least one of hashes, file_ids is required")
+        if file_service_name is not None and file_service_key is not None:
+            raise ValueError("Exactly one of file_service_name, file_service_key is required")
 
-    def undelete_files(self, hashes: abc.Iterable[str]) -> None:
-        self._api_request("POST", self._UNDELETE_FILES_PATH, json={"hashes": hashes})
+        payload: dict[str, T.Any] = {}
+        if hashes is not None:
+            payload["hashes"] = hashes
+        if file_ids is not None:
+            payload["file_ids"] = file_ids
+        if file_service_name is not None:
+            payload["file_service_name"] = file_service_name
+        if file_service_key is not None:
+            payload["file_service_key"] = file_service_key
+        if reason is not None:
+            payload["reason"] = reason
 
-    def archive_files(self, hashes: abc.Iterable[str]) -> None:
-        self._api_request("POST", self._ARCHIVE_FILES_PATH, json={"hashes": hashes})
+        self._api_request("POST", self._DELETE_FILES_PATH, json=payload)
 
-    def unarchive_files(self, hashes: abc.Iterable[str]) -> None:
-        self._api_request("POST", self._UNARCHIVE_FILES_PATH, json={"hashes": hashes})
+    def undelete_files(
+        self,
+        hashes: T.Optional[abc.Iterable[str]] = None,
+        file_ids: T.Optional[abc.Iterable[int]] = None,
+        file_service_name: T.Optional[str] = None,
+        file_service_key: T.Optional[str] = None,
+    ) -> None:
+        if hashes is None and file_ids is None:
+            raise ValueError("At least one of hashes, file_ids is required")
+        if file_service_name is not None and file_service_key is not None:
+            raise ValueError("Exactly one of file_service_name, file_service_key is required")
+
+        payload: dict[str, T.Any] = {}
+        if hashes is not None:
+            payload["hashes"] = hashes
+        if file_ids is not None:
+            payload["file_ids"] = file_ids
+        if file_service_name is not None:
+            payload["file_service_name"] = file_service_name
+        if file_service_key is not None:
+            payload["file_service_key"] = file_service_key
+
+        self._api_request("POST", self._UNDELETE_FILES_PATH, json=payload)
+
+    def archive_files(
+        self,
+        hashes: T.Optional[abc.Iterable[str]] = None,
+        file_ids: T.Optional[abc.Iterable[int]] = None
+    ) -> None:
+        if hashes is None and file_ids is None:
+            raise ValueError("At least one of hashes, file_ids is required")
+
+        payload: dict[str, T.Any] = {}
+        if hashes is not None:
+            payload["hashes"] = hashes
+        if file_ids is not None:
+            payload["file_ids"] = file_ids
+
+        self._api_request("POST", self._ARCHIVE_FILES_PATH, json=payload)
+
+    def unarchive_files(
+        self,
+        hashes: T.Optional[abc.Iterable[str]] = None,
+        file_ids: T.Optional[abc.Iterable[int]] = None
+    ) -> None:
+        if hashes is None and file_ids is None:
+            raise ValueError("At least one of hashes, file_ids is required")
+
+        payload: dict[str, T.Any] = {}
+        if hashes is not None:
+            payload["hashes"] = hashes
+        if file_ids is not None:
+            payload["file_ids"] = file_ids
+
+        self._api_request("POST", self._UNARCHIVE_FILES_PATH, json=payload)
 
     def clean_tags(self, tags: abc.Iterable[str]) -> list[str]:
         response = self._api_request("GET", self._CLEAN_TAGS_PATH, params={"tags": json.dumps(tags, cls=JSONEncoder)})
         return response.json()["tags"]
 
+    def search_tags(
+        self,
+        search: str,
+        tag_service_key: T.Optional[str] = None,
+        tag_service_name: T.Optional[str] = None
+    ) -> list[dict[str, T.Union[str, int]]]:
+        if tag_service_name is not None and tag_service_key is not None:
+            raise ValueError("Exactly one of tag_service_name, tag_service_key is required")
+
+        payload: dict[str, T.Any] = {"search": search}
+        if tag_service_key is not None:
+            payload["tag_service_key"] = tag_service_key
+        if tag_service_name is not None:
+            payload["tag_service_name"] = tag_service_name
+
+        response = self._api_request("GET", self._SEARCH_TAGS_PATH, params=payload)
+        return response.json()["tags"]
+
     def add_tags(
         self,
-        hashes: abc.Iterable[str],
+        hashes: T.Optional[abc.Iterable[str]] = None,
+        file_ids: T.Optional[abc.Iterable[int]] = None,
         service_names_to_tags: T.Optional[T.Mapping[str, abc.Iterable[str]]] = None,
         service_keys_to_tags: T.Optional[T.Mapping[str, abc.Iterable[str]]] = None,
         service_names_to_actions_to_tags: T.Optional[
@@ -329,18 +425,24 @@ class Client:
             T.Mapping[str, T.Mapping[T.Union[int, TagAction], abc.Iterable[str]]]
         ] = None,
     ) -> None:
+        if hashes is None and file_ids is None:
+            raise ValueError("At least one of hashes, file_ids is required")
         if (
             service_names_to_tags is None
             and service_keys_to_tags is None
             and service_names_to_actions_to_tags is None
-            and service_keys_to_tags is None
+            and service_keys_to_actions_to_tags is None
         ):
             raise ValueError(
                 "At least one of service_names_to_tags, service_keys_to_tags, service_names_to_actions_to_tags or "
                 "service_keys_to_actions_to_tags is required"
             )
 
-        payload: dict[str, T.Any] = {"hashes": hashes}
+        payload: dict[str, T.Any] = {}
+        if hashes is not None:
+            payload["hashes"] = hashes
+        if file_ids is not None:
+            payload["file_ids"] = file_ids
         if service_names_to_tags is not None:
             payload["service_names_to_tags"] = service_names_to_tags
         if service_keys_to_tags is not None:
@@ -392,14 +494,21 @@ class Client:
 
     def associate_url(
         self,
-        hashes: abc.Iterable[str],
+        hashes: T.Optional[abc.Iterable[str]] = None,
+        file_ids: T.Optional[abc.Iterable[int]] = None,
         urls_to_add: T.Optional[abc.Iterable[str]] = None,
         urls_to_delete: T.Optional[abc.Iterable[str]] = None,
     ) -> None:
+        if hashes is None and file_ids is None:
+            raise ValueError("At least one of hashes, file_ids is required")
         if urls_to_add is None and urls_to_delete is None:
             raise ValueError("At least one of urls_to_add, urls_to_delete is required")
 
-        payload = {"hashes": hashes}
+        payload: dict[str, T.Any] = {}
+        if hashes is not None:
+            payload["hashes"] = hashes
+        if file_ids is not None:
+            payload["file_ids"] = file_ids
         if urls_to_add is not None:
             urls_to_add = urls_to_add
             payload["urls_to_add"] = urls_to_add
@@ -409,6 +518,35 @@ class Client:
 
         self._api_request("POST", self._ASSOCIATE_URL_PATH, json=payload)
 
+    def set_notes(self, notes: dict[str, str], hash_: T.Optional[str] = None, file_id: T.Optional[int] = None) -> None:
+        if (hash_ is None and file_id is None) or (hash_ is not None and file_id is not None):
+            raise ValueError("Exactly one of hash_, file_id is required")
+
+        payload: dict[str, T.Any] = {"notes": notes}
+        if hash_ is not None:
+            payload["hash"] = hash_
+        if file_id is not None:
+            payload["file_id"] = file_id
+
+        self._api_request("POST", self._SET_NOTES_PATH, json=payload)
+
+    def delete_notes(
+        self,
+        note_names: abc.Iterable[str],
+        hash_: T.Optional[str] = None,
+        file_id: T.Optional[int] = None
+    ) -> None:
+        if (hash_ is None and file_id is None) or (hash_ is not None and file_id is not None):
+            raise ValueError("Exactly one of hash_, file_id is required")
+
+        payload: dict[str, T.Any] = {"note_names": note_names}
+        if hash_ is not None:
+            payload["hash"] = hash_
+        if file_id is not None:
+            payload["file_id"] = file_id
+
+        self._api_request("POST", self._DELETE_NOTES_PATH, json=payload)
+
     def get_cookies(self, domain: str) -> list[list[T.Union[str, int]]]:
         response = self._api_request("GET", self._GET_COOKIES_PATH, params={"domain": domain})
         return response.json()["cookies"]
@@ -432,7 +570,10 @@ class Client:
         return response.json()["page_info"]
 
     def add_files_to_page(
-        self, page_key: str, file_ids: T.Optional[abc.Iterable[int]], hashes: T.Optional[abc.Iterable[str]]
+        self,
+        page_key: str,
+        file_ids: T.Optional[abc.Iterable[int]] = None,
+        hashes: T.Optional[abc.Iterable[str]] = None
     ) -> None:
         if file_ids is None and hashes is None:
             raise ValueError("At least one of file_ids, hashes is required")
@@ -457,7 +598,8 @@ class Client:
         tag_service_key: T.Optional[str] = None,
         file_sort_type: T.Optional[T.Union[int, FileSortType]] = None,
         file_sort_asc: T.Optional[bool] = None,
-    ) -> list[int]:
+        return_hashes: T.Optional[bool] = None,
+    ) -> T.Union[list[int], list[str]]:
         if file_service_name is not None and file_service_key is not None:
             raise ValueError("Exactly one of file_service_name, file_service_key is required")
         if tag_service_name is not None and tag_service_key is not None:
@@ -478,17 +620,22 @@ class Client:
             parameters["file_sort_type"] = file_sort_type
         if file_sort_asc is not None:
             parameters["file_sort_asc"] = json.dumps(file_sort_asc, cls=JSONEncoder)
+        if return_hashes is not None:
+            parameters["return_hashes"] = json.dumps(return_hashes, cls=JSONEncoder)
 
         response = self._api_request("GET", self._SEARCH_FILES_PATH, params=parameters)
-        return response.json()["file_ids"]
+        return response.json()["hashes" if return_hashes else "file_ids"]
 
     def get_file_metadata(
         self,
         hashes: T.Optional[abc.Iterable[str]] = None,
         file_ids: T.Optional[abc.Iterable[int]] = None,
+        create_new_file_ids: T.Optional[bool] = None,
         only_return_identifiers: T.Optional[bool] = None,
+        only_return_basic_information: T.Optional[bool] = None,
         detailed_url_information: T.Optional[bool] = None,
         hide_service_name_tags: T.Optional[bool] = None,
+        include_notes: T.Optional[bool] = None
     ) -> list[dict[str, T.Any]]:
         if hashes is None and file_ids is None:
             raise ValueError("At least one of hashes, file_ids is required")
@@ -499,12 +646,18 @@ class Client:
         if file_ids is not None:
             parameters["file_ids"] = json.dumps(file_ids, cls=JSONEncoder)
 
+        if create_new_file_ids is not None:
+            parameters["create_new_file_ids"] = json.dumps(create_new_file_ids, cls=JSONEncoder)
         if only_return_identifiers is not None:
             parameters["only_return_identifiers"] = json.dumps(only_return_identifiers, cls=JSONEncoder)
+        if only_return_basic_information is not None:
+            parameters["only_return_basic_information"] = json.dumps(only_return_basic_information, cls=JSONEncoder)
         if detailed_url_information is not None:
             parameters["detailed_url_information"] = json.dumps(detailed_url_information, cls=JSONEncoder)
         if hide_service_name_tags is not None:
             parameters["hide_service_name_tags"] = json.dumps(hide_service_name_tags, cls=JSONEncoder)
+        if include_notes is not None:
+            parameters["include_notes"] = json.dumps(include_notes, cls=JSONEncoder)
 
         response = self._api_request("GET", self._GET_FILE_METADATA_PATH, params=parameters)
         return response.json()["metadata"]
